/*
 * Player
 * description: abstract superclass for players in 2 player board games
 * history:
 *	2/15/93 [Erik Kay] - created
 *	6/10/93 [Erik Kay] - added PlayerState
 *	6/13/93	[Erik Kay] - removed "active" variable
 */

#import <objc/Object.h>
#import "SplatRules.h"
#import "Board.h"

typedef enum {PLAYER_STOPPED, PLAYER_PAUSED, PLAYER_ACTIVE, PLAYER_INACTIVE} PlayerState;

@interface Player:Object
{
    Rules *rules;
    Board *currentBoard;
    int pieceType, opposingPiece;
    id game;
    BOOL debug;
    char *playerName;
    PlayerState currentState;
}

- initWithRules:(Rules *)r andPieceType:(square_state)type;
- doNextMove:(Board *)b;
- (square_state)pieceType;
- setGame:g;
- gameOver:(Board *)b;
- setPlayerState:(PlayerState)state;
- (PlayerState)playerState;
- setDebug:(BOOL)state;
- (char *)playerName;

/* debugging info */
- (int)moveTime;
- (int)numLeaves;
- (int)pruningPercent;

@end

@interface Object (PlayerDelegate)
- startMove:(move *)mv;
@end
